#!/usr/bin/env python
# import
from __future__ import print_function
## batteries
import os
import sys
import pytest
## 3rd party
import numpy as np
import pandas as pd
## package
from MGSIM import Utils
from MGSIM.Commands import Communities as Communities_CMD

# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
def test_help(script_runner):
    ret = script_runner.run('MGSIM', 'communities', '-h')
    assert ret.success

def test_main(script_runner):
    genomeList = os.path.join(data_dir, 'genome_list.txt')
    output_prefix = os.path.join(data_dir, 'comm')
    ret = script_runner.run('MGSIM', 'communities',
                            genomeList, output_prefix,
                            '--n-comm', '2', '--rnd-seed', '34847')
    assert ret.success

def test_main_ncomm1(script_runner):
    genomeList = os.path.join(data_dir, 'genome_list.txt')
    output_prefix = os.path.join(data_dir, 'comm')
    ret = script_runner.run('MGSIM', 'communities',
                            genomeList, output_prefix,
                            '--n-comm', '1', '--richness', '1')
    assert ret.success




#!/usr/bin/env python
# import
from __future__ import print_function
## batteries
import os
import sys
import uuid
import pytest
import subprocess
## 3rd party
import numpy as np
import pandas as pd
## package
from MGSIM import Utils
from MGSIM import SimHtReads
from MGSIM.Commands import HtReads as HtReads_CMD

# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


def validate_fastq(data_dir):
    # validating fastq
    output_R = os.path.join(data_dir, 'TEST', '1', 'R1.fq')
    ret = subprocess.run(['fqtools', 'validate', output_R])
    assert ret.returncode == 0
    output_R = os.path.join(data_dir, 'TEST', '1', 'R2.fq')
    ret = subprocess.run(['fqtools', 'validate', output_R])
    assert ret.returncode == 0

# tests
def test_barcode_gen():
    """Testing the generation of barcodes
    """
    n_barcodes = 1000
    barcodes = SimHtReads.barcodes(n_barcodes)
    assert len(barcodes) == n_barcodes
    assert type(barcodes) is np.ndarray

def test_help(script_runner):
    ret = script_runner.run('MGSIM', 'ht_reads', '-h')
    assert ret.success

def test_main(script_runner):
    genome_table = os.path.join(data_dir, 'genome_list.txt')
    abund_table = os.path.join(data_dir, 'comm_wAbund.txt')
    temp_dir = os.path.join(data_dir, str(uuid.uuid4()))
    output_prefix = os.path.join(data_dir, 'TEST')
    
    ret = script_runner.run('MGSIM', 'ht_reads', '--art-paired',
                            '--tmp-dir', temp_dir,
                            '--barcode-total', '20',
                            '--barcode-chunks', '2',
                            '--seq-depth', '1e3',
                            '--rndSeed', '8294',
                            genome_table, abund_table, output_prefix)
    assert ret.success
    validate_fastq(data_dir)

def test_main_multi(script_runner):
    genome_table = os.path.join(data_dir, 'genome_list.txt')
    abund_table = os.path.join(data_dir, 'comm_wAbund.txt')
    temp_dir = os.path.join(data_dir, str(uuid.uuid4()))
    output_prefix = os.path.join(data_dir, 'TEST')
    
    ret = script_runner.run('MGSIM', 'ht_reads',
                            '--art-paired', '-n', '2',
                            '--tmp-dir', temp_dir,
                            '--seq-depth', '1e4',
                            '--rndSeed', '8294',
                            genome_table, abund_table, output_prefix)
    assert ret.success
    validate_fastq(data_dir)

def test_main_zeros(script_runner):
    genome_table = os.path.join(data_dir, 'genome_list.txt')
    abund_table = os.path.join(data_dir, 'comm_wAbund_zeros.txt')
    temp_dir = os.path.join(data_dir, str(uuid.uuid4()))
    output_prefix = os.path.join(data_dir, 'TEST')
    
    ret = script_runner.run('MGSIM', 'ht_reads', '--art-paired',
                            '--tmp-dir', temp_dir,
                            '--barcode-total', '20',
                            '--barcode-chunks', '2',
                            '--seq-depth', '1e3',
                            '--rndSeed', '8294',
                            genome_table, abund_table, output_prefix)
    assert ret.success
    validate_fastq(data_dir)
    
def test_main_prefix(script_runner):
    genome_table = os.path.join(data_dir, 'genome_list.txt')
    abund_table = os.path.join(data_dir, 'comm_wAbund.txt')
    temp_dir = os.path.join(data_dir, str(uuid.uuid4()))
    output_prefix = os.path.join(data_dir, 'TEST')
    
    ret = script_runner.run('MGSIM', 'ht_reads', '--art-paired',
                            '--tmp-dir', temp_dir,
                            '--barcode-total', '20',
                            '--barcode-chunks', '2',
                            '--seq-depth', '1e3',
                            '--read-name', '{readID}_BX:Z:{barcodeID}',
                            '--rndSeed', '8294',
                            genome_table, abund_table, output_prefix)
    assert ret.success
    validate_fastq(data_dir)

def test_main_large_frag_sd(script_runner):
    genome_table = os.path.join(data_dir, 'genome_list.txt')
    abund_table = os.path.join(data_dir, 'comm_wAbund.txt')
    temp_dir = os.path.join(data_dir, str(uuid.uuid4()))
    output_prefix = os.path.join(data_dir, 'TEST')
    
    ret = script_runner.run('MGSIM', 'ht_reads', '--art-paired',
                            '--tmp-dir', temp_dir,
                            '--frag-size-mean', '10000',
                            '--frag-size-sd', '9000',
                            '--barcode-total', '20',
                            '--barcode-chunks', '2',
                            '--seq-depth', '5e3',
                            genome_table, abund_table, output_prefix)
    assert ret.success
    validate_fastq(data_dir)     



#!/usr/bin/env python

"""
ht_reads: simulating haplotagging reads

Usage:
  ht_reads [options] <genome_table> <abund_table> <output_dir>
  ht_reads -h | --help
  ht_reads --version

Options:
  <genome_table>          Taxon genome info.
  <abund_table>           Taxon abundance info.
  <output_dir>            Output directory.
  --barcode-total=<bt>    Number of barcodes 
                          [Default: 1e3]
  --barcode-chunks=<bc>   Chunk size for parallel processing of barcodes
                          [Default: 1e2]
  --seq-depth=<d>         Number of (paired) Illumina reads per sample.
                          [Default: 1e5]
  --frag-size-mean=<fsm>  Mean fragment size of input gDNA (bp)
                          [Default: 10000]
  --frag-size-sd=<fss>    Fragment size standard deviation of input gDNA (bp)
                          [Default: 1000]
  --frag-size-min=<fsa>   Min fragment size of input gDNA (bp)
                          [Default: 8000]
  --frag-size-max=<fsa>   Min fragment size of input gDNA (bp)
                          [Default: 20000]
  --frag-bc-mean=<fbm>    Mean fragment-to-barcode count
                          [Default: 5]
  --frag-bc-sd=<fbs>      Standard deviation of fragment-to-barcode count
                          [Default: 1]
  --art-paired            art_illumina "paired"  parameter.
  --art-len=<al>          art_illumina "len" parameter.
                          [Default: 150] 
  --art-mflen=<am>        art_illumina "mflen" parameter.
                          Use 0 to turn off
                          [Default: 200]
  --art-sdev=<ad>         art_illumina "sdev" parameter.
                          [Default: 10]
  --art-seqSys=<as>       art_iilumina "seqSys" parameter.
                          [Default: HS25]
  --rndSeed=<rs>          Random Seed for Art. If None, then randomly set.
                          [Default: None]
  --tmp-dir=<td>          Temporary directory. 
                          [Default: .sim_reads]
  --read-name=<bft>       Read name format
                          [Default: {readID} BX:Z:{barcodeID}]
  -n=<n>                  Number of cpus. 
                          [Default: 1]
  --debug                 Debug mode (no subprocesses verbose output).
  -h --help               Show this screen.
  --version               Show version.

Description:
  Simulating reads for each taxon in each synthetic community

  abund_table
  -----------
  * tab-delimited
  * must contain 3 columns
    * "Community" = community ID (ie., sample ID)
    * "Taxon" = taxon name
    * "Perc_rel_abund" = percent relative abundance of the taxon
  * NOTE: For ht_reads, only the first community will be used (no multi-samples)

  genome_table
  ------------
  * tab-delimited
  * must contain 2 columns
    * "Taxon" = taxon name
    * "Fasta" = genome fasta file path
  * other columns are allowed

  Output
  ------
  * tsv file of simulated genome fragments
    * These fragments were used as reference for read simulation
  * Read1 fastq [& Read2 fastq]
    * Default read header naming: <fragment UUID> BX:Z:<barcode_ID>
"""

# import
## batteries
from docopt import docopt
import sys,os
import re
import time
import logging
from pprint import pprint
from shutil import rmtree
from functools import partial
import multiprocessing as mp
from distutils.spawn import find_executable
## 3rd party
import numpy as np
## application
from MGSIM import SimReads
from MGSIM import SimHtReads
## logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)

def _flatten(list_of_lists):
    flattened_list = []
    for x in list_of_lists:
        for y in x:
            flattened_list.append(y)
    return flattened_list

def flatten(list_of_lists):
    flattened_list1 = []
    flattened_list2 = []
    for x in list_of_lists:
        for i,y in enumerate(x):
            if i % 2 == 0:
                flattened_list1.append(y)
            else:
                flattened_list2.append(y)
    return [_flatten(flattened_list1),
            _flatten(flattened_list2)]

def main(args):
    """ Main interface
    """
    # check executables
    exe = 'art_illumina'
    if find_executable(exe) == '':
        raise IOError('Cannot find executable: {}'.format(exe))
    
    # load tables
    ## table mapping taxa to genomes
    genome_table = SimReads.load_genome_table(args['<genome_table>'])
    ## table of taxon relative abundances
    abund_table = SimReads.load_abund_table(args['<abund_table>'])

    # Sim Params
    ## random seed
    if args['--rndSeed'] is None or args['--rndSeed'] == 'None':
        rndSeed = None
    else:
        rndSeed = int(args['--rndSeed'])

    # simulating per community
    comms = abund_table.Community.unique()
    for comm in comms:
        logging.info('Simulating reads for community: {}'.format(comm))
        comm_tbl = abund_table.loc[abund_table.Community == comm,]
        sim_per_community(args, comm_id = comm,
                          abund_table = comm_tbl,
                          genome_table = genome_table,
                          rndSeed = rndSeed)


def sim_per_community(args, comm_id, abund_table, genome_table, rndSeed):
    # I/O
    outdir = os.path.join(args['<output_dir>'], str(comm_id))
    ## temp dir
    tmp_dir = os.path.join(args['--tmp-dir'], str(comm_id))
    if not os.path.isdir(tmp_dir):
        os.makedirs(tmp_dir)
    
    ## batching by barcodes
    ### creating barcode ID array
    n_barcodes = int(float(args['--barcode-total']))
    barcodes = SimHtReads.barcodes(n_barcodes)
    ### simulating barcodes
    func = partial(SimHtReads.sim_barcode_reads,
                   genome_table=genome_table,
                   abund_table=abund_table,
                   tmp_dir=tmp_dir,
                   args=args,
                   rndSeed=rndSeed)
    barcodes = SimHtReads.bin_barcodes(barcodes,args['--barcode-chunks'])
    msg = 'Processing barcodes in {} chuncks of {} barcodes each'
    logging.info(msg.format(len(barcodes), args['--barcode-chunks']))
    if args['--debug'] or args['-n'] < 2:
        files = map(func, barcodes)
    else:
        Pool = mp.Pool(args['-n']) 
        files = Pool.map(func, barcodes)
        Pool.close()
    # flatten batched output
    (fq_files,tsv_files) = flatten(files)
    # combining all frag tsv
    SimHtReads.combine_frag_tsv(tsv_files, outdir)
    # combining all reads
    SimHtReads.combine_reads(fq_files, outdir, name_fmt=args['--read-name'],
                             seq_depth=int(float(args['--seq-depth'])))
    # removing temp directory    
    if args['--debug'] is False:
        logging.info('Removing temporary directory: {}'.format(args['--tmp-dir']))
        rmtree(args['--tmp-dir'], ignore_errors=True)
        if os.path.isdir(args['--tmp-dir']):
            msg = 'Could not remove directory: {}'
            logging.warning(msg.format(args['--tmp-dir']))
            
def opt_parse(args=None):
    if args is None:        
        args = docopt(__doc__, version='0.1')
    else:
        args = docopt(__doc__, version='0.1', argv=args)
    args['-n'] = int(args['-n'])
    os.environ['OMP_NUM_THREADS'] = '1'
    main(args)
   



#!/usr/bin/env python
# import
from __future__ import print_function
## batteries
import os
import sys
import pytest
## 3rd party
import pandas as pd
## package
from MGSIM import Utils
from MGSIM.Commands import Genome_download as Genome_download_CMD

# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
def test_help(script_runner):
    ret = script_runner.run('MGSIM', 'genome_download', '-h')
    assert ret.success

def test_main(script_runner):
    acc_tbl = os.path.join(data_dir, 'genome_download.tsv')
    ret = script_runner.run('MGSIM', 'genome_download',
                            '--debug', '-d', data_dir, acc_tbl)
    assert ret.success
    
def test_main_rename(script_runner):
    acc_tbl = os.path.join(data_dir, 'genome_download.tsv')
    ret = script_runner.run('MGSIM', 'genome_download',
                            '--debug', '-r', '-d', data_dir, acc_tbl)
    assert ret.success

def test_main_badAcc(script_runner):
    acc_tbl = os.path.join(data_dir, 'genome_download_badAcc.tsv')
    ret = script_runner.run('MGSIM', 'genome_download',
                            '--debug', '-r', '-d', data_dir, acc_tbl)
    assert ret.success

def test_main_dupTaxa(script_runner):
    acc_tbl = os.path.join(data_dir, 'genome_download_dupTaxa.tsv')
    ret = script_runner.run('MGSIM', 'genome_download',
                            '--debug', '-r', '-d', data_dir, acc_tbl)
    assert not ret.success



from __future__ import print_function
# import
## batteries
import sys, os
import numpy as np
import pandas as pd
try:
    from StringIO import StringIO
except ModuleNotFoundError:
    from io import StringIO
import random
import re
import logging
from functools import partial
from itertools import chain,combinations
from operator import itemgetter
from collections import OrderedDict,defaultdict
## 3rd party
import scipy.stats as stats
from scipy.spatial import distance
from configobj import ConfigObj, flatten_errors
from Bio import SeqIO


# utility functions
def str2dict(s):
    """Parsing string (format: 'item:value,item:value')
    to create a dict object.    
    """    
    if hasattr(s, 'split'):
        l = re.split('[:,]', s)
        try:
            return {k.lower():float(v) for k,v in zip(l[0::2],l[1::2])}
        except TypeError:
            msg = 'distribution parameter values must be ints or floats.'
            raise TypeError(msg)
    else:
        return s
                
def random_insert_seq(l, seq):
    """Insert seq items at random locations in list.

    Paramters
    ---------
    l : list
        target list object
    seq : iteralbe
        items to insert in the list

    Returns
    -------
    list : target list with `seq` values randomly inserted
    """
    insert_locs = random.sample(range(len(l) + len(seq)), len(seq))
    inserts = dict(zip(insert_locs, seq))
    inputs = iter(l)
    return [inserts[pos] if pos in inserts else next(inputs)
            for pos in range(len(l) + len(seq))]

def power_neg(*args, **kwargs):
    return 1 - np.random.power(*args, **kwargs)

    
    
class _Comm(object):
    """Parent class for other classes in the module.
    """
    def __init__(self):
        pass
    
    # property/setter
    @property
    def abund_dist(self):
        return self._abund_dist
    @abund_dist.setter
    def abund_dist(self, x):
        self._abund_dist = str(x)
    @property
    def richness(self):
        return self._richness
    @richness.setter
    def richness(self, x):
        x = float(x)
        if x <= 1:
            # x = fraction of total taxa pool
            # setting x as that fraction number of taxa
            x = len(self.taxon_pool) * x
        self._richness = int(round(x,0))
        if self._richness < 1:
            self._richness = 1


class SimComms(_Comm):
    """Class for simulating taxon count data of communities.
    """
    def __init__(self, taxon_list, perm_perc, shared_perc,
                 richness, abund_dist, abund_dist_params,
                 n_comm, config=None, rnd_seed=None,
                 *args, **kwargs):
        """
        Parameters
        ----------
        See gradientComms
        """
        _Comm.__init__(self, *args, **kwargs)

        self._load_taxon_list(taxon_list)        
        self.perm_perc = perm_perc
        self.shared_perc = shared_perc
        self.richness = richness
        self.abund_dist = abund_dist
        self.abund_dist_params = str2dict(abund_dist_params)
        self.config = config
        self.n_comm = n_comm
        self.rnd_seed = rnd_seed

        if not self.rnd_seed is None and self.rnd_seed != 'None':
            random.seed(int(self.rnd_seed))
        
        # loading config; setting community parameters
        if config is not None:
            self.comm_params = self._load_config()
        else:
            self.comm_params = dict()
        self._set_comm_params()

        # lowering comm richness if taxon pool is limiting
        ## otherwise, shared_perc option throws and error
        if self.shared_perc < 100:
            self._lower_richness()

        # shared taxa
        self._set_shared_taxa()
                            
    def _get_configspec(self, strIO=True):
        """Return configspec set for instance.
        Parameters
        ----------
        strIO : bool
            return configspec as a StringIO instance
        Returns
        -------
        configspec object 
        """
        configspec = """
        [__many__]
            richness = float(0,inf, default=None)        
            abund_dist = string(default='exponential,1,0.5')
            start = float(default=None)
            end = float(default=None)
            loc = float(default=None)
            scale = float(default=None)
            sigma = float(default=None)
        """
        
        if strIO == True:
            return StringIO(configspec)
        else:
            return configspec
 
    def _load_config(self):    
        assert hasattr(self, 'config'), "No config attribute found."

        configspec = self._get_configspec()
        return ConfigObj(self.config, configspec=configspec)        
        
    def _set_comm_params(self):
        """Setting community-specific params including applying global params.
        """
        # adding to comm params if not enough set by config
        n_config_comms = len(self.comm_params.keys())
        n_diff = self.n_comm - n_config_comms
        
        for i in range(n_diff):
            self.comm_params[str(n_config_comms + i + 1)] = dict()
        
        for k,v in self.comm_params.items():
            # checking for params
            if ('richness' not in v.keys() or
                v['richness'] is None):
                v['richness'] = self.richness
            if ('abund_dist' not in v.keys() or
                v['abund_dist'] is None):
                v['abund_dist'] = self.abund_dist
            if ('abund_dist_p' not in v.keys() or
                v['abund_dist_p'] is None):
                v['abund_dist_p'] = self.abund_dist_params
            v['abund_dist_p'] = str2dict(v['abund_dist_p'])
                
    def _set_shared_taxa(self):
        """A list of taxa shared among all communities.
        The taxon list (pool) is reduced to just unshared taxa.
        """
        self.shared_taxa = self._drawFromTaxonPool(self.n_shared)
        
    def _load_taxon_list(self, fileName):
        """Loading taxon list file. Taxa order is randomly shuffled.
        
        Parameters
        ----------
        fileName : str
             name of taxon file
        """
        # load file
        if fileName == '-':
            inF = sys.stdin
        else:
            inF = open(fileName)
        df = pd.read_csv(inF, sep='\t')
        ## check headers
        diff = set(['Taxon','Fasta']) - set(df.columns.values)
        if len(diff) > 0:
            diff = ','.join(diff)
            raise ValueError('Cannot find table columns:{}'.format(diff))

        # setting taxon_pool
        self.taxon_pool = df['Taxon']
        random.shuffle(self.taxon_pool)

        # setting genome fasta dict {Taxon : Fasta}
        self.genome_fasta = {}
        for i,x in df.iterrows():
            self.genome_fasta[x['Taxon']] = x['Fasta']
        
    def _drawFromTaxonPool(self, n):
        """Draw from taxon pool, returning n-taxa;
        those taxa are removed from the pool.

        Parameters
        ----------
        n : int
            number of taxa to draw

        Returns
        -------
        list : [taxon_name1, taxon_nameN, ...]
        """
        assert n <= len(self.taxon_pool), \
            'Cannot draw {} taxa from taxon pool'.format(n)
        taxa = self.taxon_pool[:n]
        self.taxon_pool = self.taxon_pool[n:]
        return taxa


    def _lower_richness(self):
        """Lowering the richness of each community if the number of
        unique taxa (un-shared) + shared taxa is greater than the taxon
        pool from which to draw taxa.
        """
        rich = []
        for k,v in self.comm_params.items():
            try:
                rich.append(v['richness'])
            except KeyError:
                msg = 'Cannot find "richness" attribute for Comm {}'
                raise KeyError(msg.format(k))
        n_unique = np.sum([x - self.n_shared for x in rich])
        n_taxa_pool = len(self.taxon_pool)
        n_comm = len(rich)
        n_less = 0
        if n_unique + self.n_shared > n_taxa_pool:
            n_less = n_unique + self.n_shared - n_taxa_pool
            n_less = np.ceil(n_less/ n_comm) 
            n_less = int(n_less) + 1
        else:
            return 0
        for k,v in self.comm_params.items():
            new_rich = v['richness'] - n_less
            msg = 'WARNING: lowering richness ({} -> {}) for Community {}\n' + \
                  '  because the taxon pool is not large enough for the\n' + \
                  '  amount of un-shared taxa (set by --shared_perc)\n'
            sys.stderr.write(msg.format(v['richness'], new_rich, k))
            v['richness'] = new_rich
        self._n_shared = None
               
    def make_comm(self, comm_id):
        """Make a Comm object.
        
        Parameters
        ----------
        comm_id : str
            Community name from comm_params attrib
        """
        # assertions
        comm_id = str(comm_id)
        if not hasattr(self, 'comms'):
            self.comms = OrderedDict() #dict()
            
        try:
            self.comm_params[comm_id]
        except KeyError:
            raise KeyError('Cannot find community ID "{}" in '\
                           'community params\n'.format(comm_id))
        
        # init comm objects
        self.comms[comm_id] = Comm(comm_id, self)
        
    def write_comm_table(self, out_file=None, Long=True):
        """Joining comm objects into 1 dataframe and printing.
        Writing table to STDOUT.

        Parameters
        ----------
        out_file : str
            Output file name
        Long : bool
            Write table in long format
        """
        df =  pd.concat([x.taxa for x in self.values()],
                        axis=1, sort=False)

        write_index = True
        df.columns = self.keys()
        if Long == True:
            write_index = False
            # melting
            val_vars = list(df.columns)
            df['taxon'] = df.index
            df = pd.melt(df, id_vars=['taxon'], value_vars=val_vars)
            # ordering columns
            df.columns = ['taxon_name', 'library', 'rel_abund_perc']            
            # sorting 
            df.sort_values(by=['library','rel_abund_perc'], 
                           axis=0, ascending=[True,False], inplace=True)
            # converting any NAs to zeros
            df.fillna(0, inplace=True)
            # tidy taxon names
            func = lambda x: re.sub(r'[()\/:;, ]+', '_', x)
            df['taxon_name'] = df['taxon_name'].astype(str).apply(func)
            # getting rank by community (grouping by community)
            df['rank'] = df.groupby(['library'])['rel_abund_perc']\
                           .rank(method='first',ascending=False).astype('int')
            df = df[['library','taxon_name','rel_abund_perc','rank']]
            # renaming
            df.rename(columns={'library' : 'Community',
                       'taxon_name' : 'Taxon',
                       'rel_abund_perc' : 'Perc_rel_abund',
                       'rank' : 'Rank'}, inplace=True)
                        
        # writing dataframe
        if out_file is None:
            df.to_csv(sys.stdout, sep='\t', na_rep=0,
                      float_format='%.9f', index=write_index)
        else:
            df.to_csv(out_file, sep='\t', na_rep=0,
                      float_format='%.9f', index=write_index)            

    def weighted_abundances(self):
        """Calculate weighted abundance: rel_abund * genome_size
        
        Parameters
        ----------
        """
        self._set_genome_sizes()

        self.w_abund_tbl = {}
        for comm_id,comm in self.items():
            w_abunds = {}
            for x in comm.taxa.index:
                w_abunds[x] = comm.taxa[x] * self.genome_sizes[x]
            sum_abunds = sum(w_abunds.values())
            w_abunds = {k : v / sum_abunds * 100.0 for k,v in w_abunds.items()}
            self.comms[comm_id].taxa = pd.Series(w_abunds)
            
    def _set_genome_sizes(self):
        """Get the total bp for the genome sequence of all genomes
        
        Parameters
        ----------
        """
        self.genome_sizes = {}
        for taxon,fasta in self.genome_fasta.items():
            self.genome_sizes[taxon] = 0
            for record in SeqIO.parse(fasta, "fasta"):
                self.genome_sizes[taxon] += len(record.seq)        

    def beta_diversity(self, measures=['braycurtis'], outfile=None):
        """Calculate weighted abundance: rel_abund * genome_size
        
        Parameters
        ----------
        measure : list
          which scipy.spatial.distance function to use
        outfile : str
          if not None, then file written to that path
        """
        # taxon abudance table
        df =  pd.concat([x.taxa for x in self.values()],
                        axis=1, sort=False)
        df.fillna(0, inplace=True)
        # distance 
        D = defaultdict(dict)
        psbl_msr = {'braycurtis' : distance.braycurtis,
                    'jaccard' : distance.jaccard,
                    'euclidean' : distance.euclidean}
        for c1,c2 in combinations(df.columns, 2):
            for m,func in psbl_msr.items():
                if m in measures:
                    if m == 'jaccard':
                        D[m][(c1,c2)] = func(df[c1].astype(bool),
                                             df[c2].astype(bool))
                    else:
                        D[m][(c1,c2)] = func(df[c1], df[c2])
        if outfile is not None:
            with open(outfile, 'w') as outF:
                h = ['commX', 'commY', 'measure', 'distance']
                outF.write('\t'.join(h) + '\n')
                for m,v in D.items():                    
                    for c,d in v.items():
                        d = str(round(d,6))
                        c = [str(x) for x in list(c)]
                        outF.write('\t'.join(c + [m, d]) + '\n')
        return D                          
                
    @staticmethod
    def permute(comm, perm_perc):
        """Permute a certain percentage of the taxa abundances.
        Permuting just the indices of the series objects.
        In-place edit of comm table

        Parameters
        ----------
        comm : comm table object
        perm_perc : float
            percent of taxa to permute
        """
        # assertions
        perm_perc = float(perm_perc)
        assert (perm_perc >= 0 and perm_perc <= 100),\
            'perm_perc is not in range [0,100]'
        assert hasattr(comm, 'taxa'), \
            'No "taxa" attribute for comm {}'.format(comm.comm_id)

        # variables
        n_perm = int(round(perm_perc / 100 * comm.n_taxa,0))
        
        # permuting index of comm
        perm_idx = random.sample(range(comm.n_taxa), n_perm)
        perm_ig = itemgetter(perm_idx)
        n_perm_idx = set(range(comm.n_taxa)) - set(perm_idx)
                
        if len(n_perm_idx) > 0:
            n_perm_ig = itemgetter(*n_perm_idx)        
            # altering pandas series of taxa & abundances
            comm.taxa.index = random_insert_seq(n_perm_ig(comm.taxa.index),
                                                perm_ig(comm.taxa.index))
        else:
            # altering pandas series of taxa & abundances
            comm.taxa.index = random_insert_seq([],
                                                perm_ig(comm.taxa.index))

    # dict functions
    def items(self):
        return self.comms.items()
    def keys(self):
        try:
            return self.comms.keys()
        except AttributeError:
            return np.sort(list(self.comm_params.keys()))
    def values(self):
        return self.comms.values()
    
    @property
    def n_comm(self):        
        return self._n_comm
    @n_comm.setter
    def n_comm(self, x):
        try:
            self._n_comm = int(x)
        except ValueError:
            raise ValueError('n_comm must be an integer')

    @property
    def perm_perc(self):
        return self._perm_perc
    @perm_perc.setter
    def perm_perc(self, x):
        x = float(x)
        assert (x >= 0 and x <= 100), 'shared_perc must be in range 0-100'
        self._perm_perc = x        

    @property
    def shared_perc(self):
        return self._shared_perc
    @shared_perc.setter
    def shared_perc(self, x):
        x = float(x)
        assert (x >= 0 and x <= 100), 'shared_perc must be in range 0-100'
        self._shared_perc = x

    @property
    def min_richness(self):
        """The minimum richness of any community as defined by comm_params.
        """
        if not hasattr(self, '_min_richness'):
            setattr(self, '_min_richness', None)
        if self._min_richness is None:
            richness_vals = []
            for k,v in self.comm_params.items():
                try:
                    richness_vals.append(int(v['richness']))
                except KeyError:
                    raise KeyError('Cannot find richness attribute for '\
                                   'comm_id "{}"'.format(k))
            self._min_richness = min(richness_vals)
        return self._min_richness

    @property
    def n_shared(self):
        """The number of taxa that should be shared;
        defined by shared_perc * min richness of any community.
        """
        if not hasattr(self, '_n_shared'):
            setattr(self, '_n_shared', None)
        if self._n_shared is None:
            self._n_shared = self.min_richness * (self.shared_perc / 100.0)
            self._n_shared = int(round(self._n_shared,0))
        return self._n_shared

    @property
    def n_taxa_remaining(self):
        """The number of taxa that remain in taxa pool.
        """
        if not hasattr(self, '_n_taxa_remaining'):
            setattr(self, '_n_taxa_remaining', None)
        return len(self.taxon_pool)
                    
    
class Comm(_Comm):
    """Community class"""

    def __init__(self, comm_id, Comms, *args, **kwargs): 
        """
        Parameters
        ----------
        comm_id : str
            community ID
        Comms : community object
        """
        _Comm.__init__(self, *args, **kwargs)
        self.comm_id = comm_id                                                     
        self.params = Comms.comm_params[comm_id]
        self.n_shared = Comms.n_shared
        self.taxon_pool = Comms.taxon_pool
        self.richness = self.params['richness']

        # assertions
        if self.richness > self.n_shared + Comms.n_taxa_remaining:
            sys.exit('ERROR: Comm_ID {}\n'\
            '  Community richness is set too high! It is > taxon pool.\n'\
            '  There is not enough taxa to make the desired communities.\n' \
            '  You must reduce richness or increase perc_shared.\n'\
            '  NOTE: shared_perc is based on the community with the min. ' \
            'richness.\n'.format(comm_id))
        
        # selecting additional taxa beyond those shared by all comms
        ## unique taxa inserted rand in list while keeping shared taxa r-abund
        n_unique = self.richness - Comms.n_shared
        assert n_unique >= 0, 'ERROR: Comm_ID {}: the number ' \
            'of unique taxa is < 0'.format(comm_id)
        self.taxa = random_insert_seq(Comms.shared_taxa,
                                      Comms._drawFromTaxonPool(n_unique))

        
        # drawing relative abundances from the user-defined distribution
        abund_dist = self._get_abund_dist(self.params['abund_dist'],
                                          self.params['abund_dist_p'])        
        rel_abunds = abund_dist(size=self.n_taxa)
        rel_abunds = np.sort(rel_abunds / sum(rel_abunds) * 100)[::-1]
    
        # making a series for the taxa
        self.taxa = pd.Series(rel_abunds, index=self.taxa)        

    def __repr__(self):
        return self.taxa.__repr__()
        
    def _get_abund_dist(self, dist, params):
        try:
            distFunc = getattr(np.random, dist)
        except AttributeError:
            msg = 'Distribution "{}" is not supported'.format(dist)

        if dist == 'power':
            distFunc = power_neg
                        
        try:
            return partial(distFunc, **params)
        except TypeError:
            param_str = [':'.join([str(k),str(v)]) for k,v in params.items()]
            param_str = ','.join(param_str)
            msg = 'Params "{}" do not work with function "{}"'\
                  .format(param_str, dist)
            raise TypeError(msg)

    @property
    def n_taxa(self):
        return len(self.taxa)



#!/usr/bin/env python
# import
from __future__ import print_function
## batteries
import os
import sys
import pytest
## 3rd party
import numpy as np
import pandas as pd
## package
from MGSIM import Utils

# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
def test_file_status():
    # real file
    f = os.path.join(data_dir, 'Methanosarcina_barkeri_MS.fna')
    Utils.is_file(f)
    Utils.checkExists(f)
    Utils.checkEmpty(f)
    # fake file
    f = os.path.join(data_dir, 'DOES_NOT_EXIST')
    with pytest.raises(IOError):
        Utils.is_file(f)
    with pytest.raises(IOError):
        Utils.checkExists(f)
    



from __future__ import print_function
# import
## batteries
import os
import sys
import re
import time
import uuid
import logging
import subprocess
from glob import glob
from pprint import pprint
from shutil import rmtree
from random import shuffle
from functools import partial
from multiprocessing import Pool
from distutils.spawn import find_executable
## 3rd party
import numpy as np
import pandas as pd
import pyfastx
from Bio import SeqIO
#from pyfaidx import Fasta
from scipy.stats import truncnorm
## application
from MGSIM import Utils

# logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)


# functions
def bin_barcodes(barcodes, binsize=1000):
    """Binning barcodes into chunks
    Parameters
    ----------
    barcodes : iterable
        Iterable of barcodes
    binsize : int
        Size of bin for grouping barcodes
    
    Returns
    -------
    yields list of barcode (1 bin)
    """
    binsize = int(float(binsize))
    bins = np.digitize(np.arange(0,barcodes.shape[0]),
                       np.arange(0,barcodes.shape[0],binsize))
    return [barcodes[bins == x] for x in np.unique(bins)]

def sim_barcode_reads(barcodes, genome_table, abund_table, tmp_dir, args, rndSeed=None):
    """Simulate reads for each fragment associated with each barcode
    Parameters
    ----------
    barcodes : iterable
        Iterable of barcodes [str]
    genome_table : pd.dataframe
        Table that maps genome fasta to each taxon
    abund_table : pd.dataframe
        Table listing relative abundances of each taxon
    args: dict
        Command line args 
    rndSeed: int
        --rndSeed for ART
    """
    genome_table = genome_table.merge(abund_table,on=['Taxon'])

    # simulate fragment reads
    frag_tsv_files = []
    read_fq_files = []
    for barcode in barcodes:
        # number of fragments
        n_frags = np.random.normal(loc=float(args['--frag-bc-mean']),
                                   scale=float(args['--frag-bc-sd']))
        n_frags = int(round(n_frags, 0))
        
        # selecting reference genomes (with replacement)
        refs = select_refs(n_frags, genome_table)

        # sim fragment sizes
        refs = sim_frags(refs, n_frags,
                         frag_size_loc=float(args['--frag-size-mean']),
                         frag_size_sd=float(args['--frag-size-sd']),
                         frag_size_min=int(args['--frag-size-min']),
                         frag_size_max=int(args['--frag-size-max']))

        # parsing fragments
        func = partial(parse_frags, barcode=barcode, out_dir=tmp_dir, tmp_dir=tmp_dir)
        frag_files = [func(refs.loc[refs['Taxon'] == taxon]) for taxon in refs['Taxon'].unique()]
        
        # simulating reads
        for (fasta_file,tsv_file) in frag_files:
            art_params = {'--paired' : args['--art-paired'],
                          '--len' : args['--art-len'],
                          '--mflen' : args['--art-mflen'],
                          '--sdev' : args['--art-sdev'],
                          '--seqSys' : args['--art-seqSys']}
            if rndSeed is not None:
                art_params['--rndSeed'] = rndSeed
            fasta_files = sim_illumina(fasta_file, str(barcode),
                                       seq_depth=float(args['--seq-depth']),
                                       total_barcodes=int(float(args['--barcode-total'])),
                                       art_params=art_params,
                                       tmp_dir=tmp_dir,
                                       debug=args['--debug'])
            read_fq_files.append(fasta_files)
            frag_tsv_files.append(tsv_file)

    return [read_fq_files, frag_tsv_files]

def format_params(d):
    dd = {}
    for k,v in d.items():
        if v == True:
            dd[k] = ''
        elif v == False:
            continue
        else:
            dd[k] = v
    return dd

def sim_illumina(frag_fasta, barcode, seq_depth, total_barcodes,
                 art_params, tmp_dir, debug=False):
    """Simulating illumina reads for each fragment from each barcode
    Parameters
    ----------
    frag_fasta : str
        Fasta file of simulated gDNA fragments for a particular barcode
    barcode : str
        Barcode ID
    seq_depth : int
        Total sequencing depth
    total_barcodes : int
        Total number of barcodes
    art_params : dict
        art_illumina parameters
    tmp_dir : str
        temporary directory path
    debug : bool
        debug mode 
    """    
    output_prefix = os.path.splitext(frag_fasta)[0] + '_'

    # formatting params
    art_params = format_params(art_params)
    
    # calculate fold coverage to simulate
    fold = calc_fold(frag_fasta, seq_depth, total_barcodes, art_params, tmp_dir=tmp_dir)
    
    # art_illumina command
    art_params = ' '.join(['{} {}'.format(k,v) for k,v in art_params.items()])
    cmd = 'art_illumina {art_params} --noALN --id {barcode}'
    cmd += ' -f {fold} -i {input} -o {output_prefix}'
    cmd = cmd.format(art_params=art_params,
                     barcode=barcode + '-',
                     fold=fold,
                     input=frag_fasta,
                     output_prefix=output_prefix)

    ## system call
    if debug is True:
        sys.stderr.write('CMD: ' + cmd + '\n')
    try:
        res = subprocess.run(cmd, check=True, shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        raise e
    if debug is True:
        sys.stderr.write(res.stderr.decode() + '\n')
        sys.stderr.write(res.stdout.decode() + '\n')

    # check that files have been created
    R0_file = output_prefix + '.fq'
    R1_file = output_prefix + '1.fq'
    R2_file = output_prefix + '2.fq'
    if os.path.isfile(R1_file) and os.path.isfile(R2_file):
        return [R1_file, R2_file]
    elif os.path.isfile(R0_file):
        return [R0_file]
    else:
        msg = 'Cannot find art_illumina output files!'
        raise ValueError(msg)    

def get_total_seq_len(fasta_file):
    """Simple function that uses pyfastx to quickly read in a fasta,
    and then the sum of sequence lengths is returned
    """
    x = [len(seq) for h,seq in pyfastx.Fasta(fasta_file, build_index=False)]
    return sum(x)

def calc_fold(frag_fasta, seq_depth, total_barcodes, art_params, tmp_dir):
    """Calculate fold coverage to simulate per barcode
    Parameters
    ----------
    frag_fasta : str
        Fasta file of simulated gDNA fragments for a particular barcode
    seq_depth : int
        Total sequencing depth
    total_barcodes : int
        Total number of barcodes
    art_params : dict
        art_illumina parameters
    tmp_dir : str
        temp_dir
    """
    # fold coverage
    ## total length of fragments
    total_frag_len = get_total_seq_len(frag_fasta)
    ## length of reads
    try:
        read_len = int(float(art_params['--len']))
    except KeyError:
        raise KeyError('Cannot find --len param for art_illumina')
    try:
        _ = art_params['--paired']
        is_paired = True
    except KeyError:
        is_paired = False
    read_len = read_len * 2 if is_paired else read_len
    ## fold calc
    seqs_per_barcode = seq_depth / total_barcodes
    fold  = seqs_per_barcode * read_len / total_frag_len
    # return
    return fold

def combine_frag_tsv(tsv_files, output_dir, debug=False):
    """ Concat all genome fragment tsv files
    Parameters
    ----------
    tsv_files : iterable
        Iterable of fragment tsv files
    output_dir : str
        Final output directory for read files
    debug : bool
        Debug mode
    """
    # writing files
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    out_file = os.path.join(output_dir, 'fragments.tsv')
    ## header
    with open(out_file, 'w') as outF:
        outF.write('\t'.join(['Barcode', 'Frag_ID',
                              'Genome', 'Contig',
                              'Frag_start', 'Frag_end']) + '\n')
    ## body
    for F in list(set(tsv_files)):
        with open(F, 'r') as inF, open(out_file, 'a') as outF:
            for line in inF:
                outF.write(line)
        # deleting temp file
        try:
            os.unlink(F)
        except OSError:
            logging.warning('Could not remove temp file: {}'.format(F))
    logging.info('File written: {}'.format(out_file))

def combine_reads(fq_files, output_dir, name_fmt='{readID} BX:Z:{barcodeID}',
                  seq_depth=None, debug=False):
    """ Concat all read files
    Parameters
    ----------
    fq_files : iterable
        Iterable of fastq read files
    output_dir : str
        Final output directory for read files
    name_fmt : str
        How to format the read name. "{readID} BX:Z:{barcodeID}" will generate names such as "@ST-J00101:121:HYCGGBBXX:5:1101:30533:1191 BX:Z:A58B91C07D86"
    debug : bool
        Debug mode
    """    
    # split by read pairs
    R1_files = [x[0] for x in fq_files]
    try:
        R2_files = [x[1] for x in fq_files]
    except IndexError:
        R2_files = None
        
    # writing files
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    ## read1
    R1_files = _combine_reads(R1_files, output_dir, 'R1.fq', name_fmt, seq_depth)
    ## read2
    if R2_files is not None:
        R2_files = _combine_reads(R2_files, output_dir, 'R2.fq', name_fmt, seq_depth)

def fix_nucs(seq):
    ambig = {'A', 'T', 'G', 'C'}
    seq = ['N' if x not in ambig else x for x in seq.rstrip()]
    return ''.join(seq) + '\n'
        
def _combine_reads(read_files, out_dir, out_file, name_fmt, seq_depth=None):
    """Combining temporary read files.
    eg., @ST-J00101:121:HYCGGBBXX:5:1101:30533:1191 BX:Z:A58B91C07D86
    """
    if seq_depth is not None:
        cnt = 0
    out_file = os.path.join(out_dir, out_file)
    with open(out_file, 'w') as outF:
        for i,F in enumerate(read_files):
            if seq_depth is not None and cnt > seq_depth:
                break
            with open(F, 'r') as inF:
                seq_len = 0
                qual_len = 0
                for ii,line in enumerate(inF):
                    if ii % 4 == 0:   # read header
                        if seq_depth is not None:
                            cnt += 1
                            if cnt > seq_depth:
                                break
                        line = line.split('-')
                        X = name_fmt.format(readID=line[0], barcodeID=line[1])
                        outF.write(X + '\n')
                        seq_len = 0
                        qual_len = 0
                        continue
                    elif ii % 4 == 1:  # read sequence
                        seq_len = len(line.rstrip())
                        line = fix_nucs(line)
                    elif ii % 4 == 3:  # read quality
                        qual_len = len(line.rstrip())
                        if seq_len != qual_len:
                            msg = 'Line {} in file {}: read-len != qual-len'
                            raise ValueError(msg.format(ii, F))
                    outF.write(line)
            # deleting temp file
            try:
                os.unlink(F)
            except OSError:
                logging.warning('Could not remove temp file: {}'.format(F))
                        
    logging.info('File written: {}'.format(out_file))
        
def select_refs(n_frags, genome_table):
    """Selecting genomes (with replacement)
    -- Columns in genome_table --
    "Fasta" = genome file
    "Taxon" = taxon to select
    "Perc_rel_abund" = weight
    """
    probs = genome_table.Perc_rel_abund / np.sum(genome_table.Perc_rel_abund)
    refs = np.random.choice(np.arange(genome_table.shape[0]),
                            size=n_frags,
                            replace=True,
                            p=probs)
    return genome_table.loc[refs]
    
def sim_frags(refs, n_frags, frag_size_loc, frag_size_sd,
              frag_size_min, frag_size_max):
    """Simulating fragment sizes
    Parameters
    ----------
    refs : pd.dataframe
        Dataframe of reference taxa
    n_frags : int
        Number of fragments
    frag_size_loc : float
        Mean fragment size
    frag_size_sd : float
        Stdev fragment size
    frag_size_min : float
        Min fragmen size (trun-norm distribution)
    frag_size_max : float
        Max fragmen size (trun-norm distribution)
    """
    a = (frag_size_min - frag_size_loc) / frag_size_sd  # fraction of range smaller than loc
    b = (frag_size_max - frag_size_loc) / frag_size_sd  # fraction of range greater than loc
    refs.loc[:,'Frag_size'] = truncnorm.rvs(a, b,
                                            size=n_frags,
                                            loc=frag_size_loc,
                                            scale=frag_size_sd)
    return refs

def read_fasta(fasta_file):
    """Fast reading into memory of fasta file
    return: dict
    """
    seqs = {h:seq for h,seq in pyfastx.Fasta(fasta_file, build_index=False)}
    return seqs
    
def parse_frags(refs, barcode, out_dir, tmp_dir):
    """Parsing fragment from a genome and writing them to a file.
    Giving each simulated fragment a UUID. 

    Parameters
    ----------
    refs : pd.dataframe
        Dataframe of reference taxa
    barcode : str
        Barcode ID
    out_dir : str
        Output directory
    """
    regex = re.compile(r'[ ;:,(){|]+')
    # fasta file of genome
    taxon = refs['Taxon'].unique()[0]
    taxon = regex.sub('_', taxon)
    fasta_file = refs['Fasta'].unique()[0]
    # loading fasta
    f = read_fasta(fasta_file)
    contig_ids = list(f.keys())
    for x in contig_ids:
        assert('|' not in x)
    # output files
    prefix = '{}__{}'.format(barcode, taxon)
    out_file_fasta = os.path.join(out_dir, prefix + '.fasta')
    out_file_tsv = os.path.join(out_dir, prefix + '.tsv')
     
    contigs = []
    with open(out_file_fasta, 'w') as outF, open(out_file_tsv, 'w') as outT:
        # each frag to simulate for target barcode
        for idx,row in refs.iterrows():
            frag_size = int(row['Frag_size'])
            shuffle(contig_ids)
            for contig_id in contig_ids:
                contig_len = len(f[contig_id])
                if contig_len >= frag_size:
                    frag_uuid = str(uuid.uuid4()).replace('-', '')
                    frag_max_end = contig_len - frag_size
                    frag_max_end = 1 if frag_max_end < 1 else frag_max_end
                    frag_start = np.random.randint(0, frag_max_end)
                    frag_end = frag_start + frag_size
                    ## writing sequence
                    outF.write('>{}\n{}\n'.format(frag_uuid,
                                                  f[contig_id][frag_start:frag_end]))
                    contigs.append(contig_id)
                    # writing tsv of positions
                    outT.write('\t'.join([str(barcode),
                                          str(frag_uuid),
                                          str(taxon),
                                          str(contig_id),
                                          str(frag_start),
                                          str(frag_end)]) + '\n')
                    break
                
    # assert that frag was created
    if len(contigs) < refs.shape[0]:
        msg = 'For taxon "{}", cannot find contig with length >= fragment size ({})'
        raise ValueError(taxon, msg.format(frag_size))

    return [out_file_fasta, out_file_tsv]

def barcodes(n):
    """Return an array (len=n) of barcode IDs
    # barcode naming: 
      The 12-bp long i5-barcode is composed of 5bp "A-part"
      connected to a 6bp "C-part" with a constant 1bp A-handle.
      The 13-bp long i7-barcode is composed of 6bp "D-part" 
      connected to a 6bp "B-part" with a constant 1bp C-handle.
      In total we have 96 different A, B, C, D parts = 85 million combinations.
    Parameters
    ----------
    n : int
        number of barcodes
    """
    func = np.vectorize(_barcode_ids)
    return func(np.arange(n))

def _barcode_ids(start=1, end=96):
    x = _barcode_id('A') + _barcode_id('B') + _barcode_id('C') + _barcode_id('D')
    return x

def _barcode_id(Char, start=1, end=96):
    x = Char + '{0:02d}'.format(np.random.randint(start,end,dtype='int'))
    return x



#!/usr/bin/env python

# import
## batteries
import os
import sys
## 3rd party
from docopt import docopt
#import Utils
## application
from MGSIM.Commands import Communities
from MGSIM.Commands import Genome_download
from MGSIM.Commands import Genome_rename
from MGSIM.Commands import Reads
from MGSIM.Commands import HtReads

def main(args=None):
    """Main entry point for application
    """
    if args is None:
        args = sys.argv[1:]
    
    docs = """
MGSIM: simulate metagenomes for multiple synthetic communities

Usage:
  MGSIM <command> [<args>...]
  MGSIM -l | --list
  MGSIM -h | --help
  MGSIM --version

Options:
  -l --list     List subcommands.
  -h --help     Show this screen.
  --version     Show version.

Commands:
  Use the `list` option.
Description:
  Simulate metagenomes for multiple synthetic communities.
  See the sub-command documentation for more information on features.
    """
    # arg parse
    args = docopt(docs,
                  version='0.1',
                  options_first=True)

    # dict of all subcommands
    cmds = {'communities' : Communities,
            'genome_download' : Genome_download,
            'genome_rename' : Genome_rename,
            'reads' : Reads,
            'ht_reads' : HtReads}
    
    # list subcommands
    if args['--list']:
        cmd_list = '\n'.join(sorted(cmds.keys(), key=str.lower))
        print('#-- Commands --#')
        print(cmd_list)
        exit()

    # running subcommand    
    try:
        func = cmds[args['<command>']]
    except KeyError:
        msg = 'ERROR: command "{}" does not exist'
        print(msg.format(args['<command>']))
        exit()        
    func.opt_parse(args['<args>'])

    
if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
genome_download: downloading genomes

Usage:
  genome_download [options] <accession_table>
  genome_download -h | --help
  genome_download --version

Options:
  <accessin_table>  Taxon-accession table (see Description).
                    Use '-' if from STDIN.
  -d=<d>            Output directory. [Default: .]
  -e=<e>            Email to use for NCBI queries. [Default: blank@gmail.com]
  -a=<a>            Number of ambiguous nucleotides allowed in a genome. [Default: 0]
  -n=<n>            Number of cpus. [Default: 1]
  -t=<t>            Number of tries to download genomes. [Default: 10]
  -r                Rename genome sequences based on taxon name?
  --debug           Debug mode (no multiprocessing).
  -h --help         Show this screen.
  --version         Show version.

Description:
  Taxon-accession table
  ---------------------
  * tab-delimited
  * must contain 2 columns
    * "Taxon" = taxon name
    * "Accession" = NCBI accession used for downloading 
      * Possible accessions:
         * ncbi nucleotide db
         * ncbi assembly db
         * ftp url to genome (direct download)
  * other columns are allowed

  Output
  ------
  * Genome fasta files written to the specified output directory
  * A table mapping taxa to the download genome fasta file is written to STDOUT
"""

# import
import sys,os
import logging
## batteries
from docopt import docopt
from MGSIM import Genome_Download
## logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)


# opt parse
def opt_parse(args=None):
    if args is None:        
        args = docopt(__doc__, version='0.1')
    else:
        args = docopt(__doc__, version='0.1', argv=args)
    Genome_Download.main(args)
   



from __future__ import print_function
# import
## batteries
import os
import sys
import re
import time
import logging
import subprocess
from glob import glob
from shutil import rmtree
from functools import partial
from multiprocessing import Pool
from distutils.spawn import find_executable
## 3rd party
import pandas as pd
from Bio import SeqIO
## application
from MGSIM import Utils

# logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)


# functions
def tidy_taxon_names(x):
    """Remove special characters from taxon names
    """
    x = re.sub(r'[()\/:;, ]+', '_', x)
    return x

def load_genome_table(in_file):
    """Loading genome table
    Parameters
    ----------
    in_file : str
        input file path
    """
    df = pd.read_csv(in_file, sep='\t')
    
    ## check headers
    diff = set(['Taxon','Fasta']) - set(df.columns.values)
    if len(diff) > 0:
        diff = ','.join(diff)
        raise ValueError('Cannot find table columns: {}'.format(diff))

    # getting genome sizes
    df['Genome_size'] = [_genome_size(x) for i,x in df.iterrows()]

    # tidy taxon names
    df['Taxon'] = df['Taxon'].astype(str).apply(tidy_taxon_names)
    
    return df

def _genome_size(x):
    """Get the total bp for the genome sequence of all genomes
        
    Parameters
    ----------
    x : pd.Series
       Series that includes 'Fasta' in the index
    """
    bp = 0
    for record in SeqIO.parse(x['Fasta'], 'fasta'):
        bp += len(record.seq)
    return bp

def load_abund_table(in_file):
    """Loading abundance table
    Parameters
    ----------
    in_file : str
        input file path
    """
    df = pd.read_csv(in_file, sep='\t')
    
    ## check headers
    diff = set(['Community','Taxon','Perc_rel_abund']) - set(df.columns.values)
    if len(diff) > 0:
        diff = ','.join(diff)
        raise ValueError('Cannot find table columns: {}'.format(diff))

    # tidy taxon names
    df['Taxon'] = df['Taxon'].astype(str).apply(tidy_taxon_names)

    return df

def sample_taxon_list(genome_table, abund_table):
    """Creating [sample,taxon] list of lists
    Parameters
    ----------
    genome_table : pd.DataFrame
    abund_table : pd.DataFrame
    """
    # joining tables
    df = abund_table.merge(genome_table, on=['Taxon'])
    # convert to a list of lists
    sample_taxon = []
    cols = ['Community', 'Taxon', 'Genome_size', 'Fasta', 'Perc_rel_abund']
    for i,x in df[cols].iterrows():
        sample_taxon.append(x.tolist())
    return sample_taxon

def sim_illumina(sample_taxon, output_dir, seq_depth, art_params,
                 temp_dir, nproc=1, rndSeed=None, debug=False):
    """Simulate illumina reads
    Parameters
    ----------
    sample_taxon : list
        [Community,Taxon,Genome_size,Fasta,Perc_rel_abund,Fold]
    output_dir : str
        Output director for all read files
    seq_depth : int
        Sequencing depth per sample
    art_params : dict
        Parameters provided to art_illumina
    temp_dir : str
        Temporary file directory
    nproc : int
        Number of parallel processes
    debug : bool
        Debug mode
    """
    # check that simulator exists
    exe = 'art_illumina'
    if find_executable(exe) == '':
        raise IOError('Cannot find executable: {}'.format(exe))
    
    # calculating fold per genome
    for x in sample_taxon:
        genome_size = float(x[2])
        perc_rel_abund = float(x[4])
        try:
            _ = art_params['--paired']
            paired = 2
        except KeyError:
            try:
                art_params['--mflen']
                paired = 2
            except KeyError:
                paired = 1
        read_length  = art_params['--len'] * paired
        fold = perc_rel_abund / 100.0 * seq_depth * read_length / genome_size
        x.append(fold)

    # directories
    ## temp dir
    if not os.path.isdir(temp_dir):
        os.makedirs(temp_dir)
    ## output dir
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
        
    # simulate per sample
    logging.info('Simulating reads...')
    func = partial(sim_art,
                   art_params=art_params,
                   temp_dir=temp_dir,
                   rndSeed=rndSeed,
                   debug=debug)
    if debug is True:
        fq_files = map(func, sample_taxon)
    else:
        p = Pool(nproc)
        fq_files = p.map(func, sample_taxon)
    fq_files = list(fq_files)

    # combining all reads by sample
    logging.info('Combining simulated reads by sample...')    
    comms = list(set([x[0] for x in sample_taxon]))
    func = partial(combine_reads_by_sample,
                   fq_files=fq_files,
                   temp_dir=temp_dir,
                   file_prefix='illumina',
                   output_dir=output_dir,
                   debug=debug)
    if debug is True:
        res = map(func, comms)
    else:
        p = Pool(nproc)
        res = p.map(func, comms)
    res = list(res)

    # removing temp dir
    logging.info('Removing temp directory...')
    rmtree(temp_dir)
    
    # status
    for sample_list in res:
        for file_name in sample_list:
            if file_name is not None:
                logging.info('File written: {}'.format(file_name))

def sim_art(x, art_params, temp_dir, rndSeed=None, debug=False):
    """Simulate illumina reads
    Parameters
    ----------
    x : list
        [Community,Taxon,Genome_size,Fasta,Perc_rel_abund,Fold]
    """
    community = str(x[0])
    taxon = str(x[1])
    fasta = str(x[3])
    perc_rel_abund = float(x[4])
    fold = float(x[5])
    
    # output
    ## temporary directories
    out_dir = os.path.join(temp_dir, str(community), taxon)
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)
    ## temporary files
    output_prefix = os.path.join(out_dir, 'illumina')
    
    # art command
    art_params = ' '.join(['{} {}'.format(k,v) for k,v in art_params.items()])
    cmd = 'art_illumina {art_params} --noALN -f {fold} -i {input} -o {output_prefix}'
    cmd = cmd.format(art_params=art_params,
                     fold=fold,
                     rndSeed=rndSeed,
                     input=fasta,
                     output_prefix=output_prefix)
    if rndSeed is not None:
        cmd += ' --rndSeed {}'.format(rndSeed)

    ## system call
    if debug is True:
        sys.stderr.write('CMD: ' + cmd + '\n')
    try:
        res = subprocess.run(cmd, check=True, shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        raise e
    if debug is True:
        sys.stderr.write(res.stderr.decode() + '\n')
        sys.stderr.write(res.stdout.decode() + '\n')

    # check that files have been created
    R0_file = output_prefix + '.fq'
    R1_file = output_prefix + '1.fq'
    R2_file = output_prefix + '2.fq'
    if os.path.isfile(R1_file) and os.path.isfile(R2_file):
        return [community, R1_file, R2_file]
    elif os.path.isfile(R0_file):
        return [community, R0_file]
    else:
        msg = 'Cannot find art_illumina output files!'
        raise ValueError(msg)
   
def combine_reads_by_sample(sample, fq_files, temp_dir, file_prefix, output_dir, debug=False):
    """ Concat all sample-taxon read files into per-sample read files
    Parameters
    ----------
    sample : str
        Sample ID
    temp_dir : str
        Temporary directory path
    file_prefix : str
        Output file prefix
    output_dir : str
        Output directory path
    debug : bool
        Debug mode
    """
    # sorting fastq files by read pair
    sample = str(sample)
    R1_files = [x[1] for x in fq_files if x[0] == sample]
    try:
        R2_files = [x[2] for x in fq_files if x[0] == sample]
    except IndexError:
        R2_files = None

    # writing files
    output_dir = os.path.join(output_dir, str(sample))
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    ## read1
    R1_files = _combine_reads(R1_files, output_dir, 'R1.fq')
    ## read2
    if R2_files is not None:
        R2_files = _combine_reads(R2_files, output_dir, 'R2.fq')

    return [R1_files, R2_files]
        
def _combine_reads(read_files, output_dir, output_file):
    """Combine fastq read files into 1 read file.
    Parameters
    ----------
    read_files : list
        All read files to combine
    output_dir : str
        Output directory path
    output_file : str
        Output file path
    """
    output_file = os.path.join(output_dir, output_file)
    with open(output_file, 'w') as outFH:
        for in_file in read_files:
            taxon = os.path.split(os.path.split(in_file)[0])[1]
            for i,record in enumerate(SeqIO.parse(in_file, 'fastq')):
                # renaming fastq read
                name =  '{}__SEQ{}'.format(taxon, i)
                record.id = name
                record.description = name
                SeqIO.write(record, outFH, 'fastq')
            # delete temporary file
            os.remove(in_file)

    return output_file



#!/usr/bin/env python
"""
communities: simulate taxon abundances in synthetic communities

Usage:
  communities [options] <genome_table> <prefix>
  communities -h | --help
  communities --version

Options:
  <genome_table>      Genome table (see Description)
                      Use '-' if from STDIN.
  <prefix>            Output file prefix
  --n-comm=<nc>       Number of communities to simulate.
                      [default: 1]
  --richness=<r>      The number of taxa in each library.
                      Values of 0-1 will be interpreted as a fraction of the 
                      total taxon pool.
                      [default: 1]
  --abund-dist=<a>    The statistical distribution used for selecting relative
                      abundances.
                      (see numpy.random for a list of distributions).
                      [default: lognormal]
  --abund-dist-p=<p>  Abundance distribution parameters.
                      (see numpy.random for distribution params).
                      [default: mean:10,sigma:2]
  --shared-perc=<sp>  The percent of taxa shared in each community.
                      Percent set by the community with the smallest richness.
                      Example: if smallest community is 10 taxa,
                               a shared percent of 20% = 2 shared taxa.
                      The total taxon pool must be large enough to accommodate
                      all un-shared taxa.
                      [default: 100]
  --perm-perc=<pp>    How much to vary the rank-abundances between communities. 
                      Percentage = percent of taxa to permute.
                      [default: 0]
  --beta-div=<m>...   beta-diversity measures to calculate among communities.
                      See scipy.spatial.distance for options.
                      [default: braycurtis jaccard]
  --config=<c>        Config file for setting community-specific parameters
                      (& global params).
                      Community-specific parameters can include:
                      (n_comm, richness, abund_dist)
  --rnd-seed=<rs>     Seed for reprodicible simulations.
                      [default: None]
  --debug             Debug mode
  --version           Show version.
  -h --help           Show this screen.

Description:
  Simulating taxon abundances (& the relative contribution to the DNA pool)
  for each synthetic community. 

  Community alpha- and beta-diveristy can be manipulated.

  genome_table
  ------------
  * tab-delimited
  * must contain 2 columns
    * "Taxon" = taxon name
    * "Fasta" = genome fasta file path
  * other columns are allowed

  Output
  ------
  * "PREFIX_abund.txt" = taxon relative abundances for each community
    * this is the relative number of genome copies for each taxon
  * "PREFIX_wAbund.txt" = taxon relative abundances weighted by genome size
    * this is the fraction of the DNA pool for each taxon
  * "PREFIX_beta-div.txt" = beta diversity among communities
    * see the beta-div parameter for selecting beta-diversity measures
"""

# import
## batteries
from docopt import docopt
import sys,os
import logging
## application
from MGSIM.SimComms import SimComms
## logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)

# functions
def main(uargs):
    # init
    SC = SimComms(taxon_list = uargs['<genome_table>'],
                  perm_perc = uargs['--perm-perc'],
                  shared_perc = uargs['--shared-perc'],
                  richness = uargs['--richness'],
                  abund_dist = uargs['--abund-dist'],
                  abund_dist_params = uargs['--abund-dist-p'],
                  n_comm = uargs['--n-comm'],
                  config = uargs['--config'],
                  rnd_seed = uargs['--rnd-seed'])
    
    # making communities
    for comm_id in SC.keys():        
        SC.make_comm(comm_id)

    # permuting based on permute perc
    for comm_id, comm in SC.items():
        if comm.richness > 1:
            SC.permute(comm, uargs['--perm-perc'])
            
    # writing out abundance table
    out_file = uargs['<prefix>'] + '_abund.txt'
    SC.write_comm_table(out_file)
            
    # creating abundance weighted by genome length
    SC.weighted_abundances()

    # writing out weighted abundance tables
    out_file = uargs['<prefix>'] + '_wAbund.txt'
    SC.write_comm_table(out_file)

    # beta-diversity among communities
    out_file = uargs['<prefix>'] + '_beta-div.txt'
    SC.beta_diversity(measures=uargs['--beta-div'], outfile=out_file)
    
        
def opt_parse(args=None):
    if args is None:        
        args = docopt(__doc__, version='0.1')
    else:
        args = docopt(__doc__, version='0.1', argv=args)
    main(args)
   



from Utils import _table
import sys, os
import re
import numpy as np

class CommTable(_table):
    """pandas DataFrame subclass
    """
    
    def __init__(self, *args, **kwargs):
        """Subclassing pandas dataframe. CommTable.from_[csv,etc]()
        Can be used for reading in a table.
        """
        _table.__init__(self, *args, **kwargs)


    def taxonInLib(self, taxon_name, libID):
        """Checking whether taxon in the selected library.
        Parameters
        ----------
        taxon_name : str
            taxon name
        libID : str
            library ID

        Return
        ------
        b : boolean
        """
        libID = re.sub('\D+', '', libID)
        dfSub = self.df.loc[self.df['taxon_name'] == taxon_name]
        libs = [str(x) for x in dfSub['library'].tolist()]
        return libID in libs

            
    def get_taxonAbund(self, taxon_name, libID=None, abs_abund=False):
        """Getting the abundance(s) of taxon in the comm file.

        Parameters
        ----------
        taxon_name : str
            name of taxon
        libID : str
            library ID. If None, all libraries selected.
        abs_abund : bool
             return absolute abundance instead of relative abundance

        Returns
        -------
        abund_vals : iterable 
            abundance values for the taxon
        """
        retCol = 'abs_abund' if abs_abund else 'rel_abund_perc'
        assert retCol in self.df.columns, \
            '"{}" column not found'.format(retCol)
        if libID is not None:            
            df_sub =  self.df.loc[(self.df['library'] == libID) &
                                  (self.df['taxon_name'] == taxon_name)]
        else:
            df_sub =  self.df.loc[(self.df['taxon_name'] == taxon_name)]

        return df_sub[retCol].tolist()

    
    def get_unique_libIDs(self):
        """Get all unique libIDs from the community table."""
        return self.df['library'].unique().tolist()

    def get_unique_taxon_names(self):
        """Get all unique taxon names from the community table."""
        return self.df['taxon_name'].unique()        

    def get_total_abund(self, libID=None, abs_abund=False):
        """Get total abundance of each library or all libraries"""
        retCol = 'abs_abund' if abs_abund else 'rel_abund_perc'
        if libID is not None:
            return self.df.loc[self.df['library'] == libID][retCol].sum()
        else:
            return self.df[retCol].sum()
        
    @property
    def abs_abund(self):
        """The absolute abundance of each taxon based on user-value.

        Parameters
        ----------
        abs_abund : int
            total abundance of all taxa in each library
        """
        return self.df['abs_abund']

    @abs_abund.setter
    def abs_abund(self, abs_abund):
        """rel_abund_perc stored as a percent"""
        try:
            x = self.df['rel_abund_perc']
            self.df['abs_abund'] = x / 100 * int(abs_abund)
        except KeyError:
            raise KeyError('"rel_abund_perc" column not found in comm file')

        self.df['abs_abund'] = np.round(self.df['abs_abund'].tolist(), 0)
        self.df['abs_abund'] = self.df['abs_abund'].astype(int)



#!/usr/bin/env python

"""
reads: simulating reads

Usage:
  reads [options] <genome_table> <abund_table> <output_dir>
  reads -h | --help
  reads --version

Options:
  <abund_table>       Taxon abundance info.  
  <genome_table>      Taxon genome info.
  <output_dir>        Output directory.
  --sr-seq-depth=<d>  Number of (paired) Illumina reads per sample.
                      [Default: 1e5]
  --art-paired        art_iilumina --paired parameter.
  --art-len=<al>      art_illumina --len parameter.
                      [Default: 150] 
  --art-mflen=<am>    art_illumina --mflen parameter.
                      Use 0 to turn off
                      [Default: 200]
  --art-sdev=<ad>     art_illumina --sdev parameter.
                      [Default: 10]
  --art-seqSys=<as>   art_illumina --seqSys parameter.
                      [Default: HS25]
  --tmp-dir=<td>      Temporary directory
                      [Default: .sim_reads]
  --rndSeed=<rs>      Random Seed for Art. If None, then randomly set.
                      [Default: None]
  -n=<n>              Number of cpus. 
                      [Default: 1]
  --debug             Debug mode (no subprocesses; verbose output)
  -h --help           Show this screen.
  --version           Show version.

Description:
  Simulating reads for each taxon in each synthetic community

  abund_table
  -----------
  * tab-delimited
  * must contain 3 columns
    * "Community" = community ID (ie., sample ID)
    * "Taxon" = taxon name
    * "Perc_rel_abund" = percent relative abundance of the taxon

  genome_table
  ------------
  * tab-delimited
  * must contain 2 columns
    * "Taxon" = taxon name
    * "Fasta" = genome fasta file path
  * other columns are allowed

  Output
  ------
  * A set of read files for each sample
    * directory structure: OUTPUT_DIR/COMMUNITY/read_files
    * read sequences are named by the taxon they originate from
"""

# import
## batteries
from docopt import docopt
import sys,os
import re
import logging
from functools import partial
import multiprocessing as mp
import logging
## application
from MGSIM import SimReads
## logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)


def main(args):
    # load tables
    genome_table = SimReads.load_genome_table(args['<genome_table>'])
    abund_table = SimReads.load_abund_table(args['<abund_table>'])
    # create pairwise
    sample_taxon = SimReads.sample_taxon_list(genome_table,
                                              abund_table)
    # simulating reads
    ## art params
    art_params = {'--paired' : args['--art-paired'],
                  '--len' : int(args['--art-len']),
                  '--mflen' : float(args['--art-mflen']),
                  '--sdev' : float(args['--art-sdev']),
                  '--seqSys' : args['--art-seqSys']}
    if art_params['--paired'] is True:
        art_params['--paired'] = ''
    else:
        art_params.pop('--paired', None)
    if int(art_params['--mflen']) <= 0:
        art_params.pop('--mflen', None)
    ### random seed
    if args['--rndSeed'] is None or args['--rndSeed'] == 'None':
        rndSeed = None
    else:
        rndSeed = int(args['--rndSeed'])
    ## read simulate
    SimReads.sim_illumina(sample_taxon,
                          output_dir=args['<output_dir>'],
                          seq_depth=float(args['--sr-seq-depth']),
                          art_params=art_params,
                          temp_dir=args['--tmp-dir'],
                          nproc=int(float(args['-n'])),
                          rndSeed=rndSeed,
                          debug=args['--debug'])
    
def opt_parse(args=None):
    if args is None:        
        args = docopt(__doc__, version='0.1')
    else:
        args = docopt(__doc__, version='0.1', argv=args)
    main(args)
   



"""Utility scripts for application"""

# import
## batteries
import os,sys
import re
import time
import logging
import platform
import subprocess
from pprint import pprint
from itertools import chain
from functools import partial
import random
import glob

## 3rd party
import numpy as np
import pandas as pd

# logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)

def get_os():
    """ Get operating system; only works for unix-like machines
    """
    OS = platform.uname()[0]
    if OS == 'Linux':
        OS = 'linux'
    elif OS == 'Darwin':
        OS = 'mac'
    else:
        sys.stderr.write('OS: "{}" not supported\n'.format(OS))
    return OS

def is_file(fileName):
    """ Does file exist? with custom output message
    """
    if os.path.isfile(fileName) is False:
        raise IOError('"{}" does not exist'.format(fileName))
        
def sys_call(cmd, quiet=False):
    """System call of command.
    
    Parameters
    ----------
    cmd : str
        The command to run as a system call
    quiet : bool
        Suppress the system command output

    Returns
    -------
    output : str
       system call output
    err : str
       system call error
    """
    try:
        if quiet:
            DEVNULL = open(os.devnull, 'w')
            proc = subprocess.Popen([cmd], shell=True, stdout=DEVNULL)
        else:
            proc = subprocess.Popen([cmd], shell=True)
    except subprocess.CalledProcessError:
        pass # handle errors in the called executable
    except OSError:
        raise OSError('No executable for command: "{}"\n'.format(cmd))
    output, err = proc.communicate()
    return output, err

def checkExists(f):
    """ Check that the file `f` exists."""
    if not os.path.isfile(f):
        msg = '"{}" not found. Did you provide the full PATH?'
        raise IOError(msg.format(f))

def checkEmpty(f):
    """ Check that the file `f` is not empty"""
    if os.stat(f).st_size == 0:
        msg = '"{}" is empty!'
        raise IOError(msg.format(f))

def parseGenomeList(inFile, filePath=None, check_exists=True):
    """Parsing the genome list file.

    Parameters
    ----------
    inFile : str
        file name of genome list file
    filePath : str
        The absolute path to genome sequence files. a
    check_exists : bool
        Check if genome sequence files exist
    
    Returns:
    list : [[taxonName, genomeFile], ...]
    """
    # parse file as list
    genomeList = []
    with open(inFile, 'rb') as inF:
        for line in inF:
            row = line.rstrip().split('\t')
            
            if row[0] == '' or row[1] == '':
                raise IOError("Necessary row value is empty!")

            if len(row) < 2:
                raise ValueError('Need format: "taxonName<tab>fileName";'
                                 'for row: "{}"'.format(row))
            else:
                (taxonName,fileName) = row[:2]
                
            # path to genome file
            if filePath is not None:
                fileName = os.path.join(filePath, fileName)

            # checking for file existence
            if check_exists:
                checkExists(fileName)
                
            #genomeList[fileName] = taxonName
            genomeList.append((taxonName,fileName))
                
    return genomeList


def describe_builtin(obj):
    """ Describe a builtin function if obj.__doc__
    available.

    Parameters
    ----------
    obj : python object
    
    Returns
    -------
    iterator : builtin args
    """
    #wi('+Built-in Function: %s' % obj.__name__)
    # Built-in functions cannot be inspected by
    # inspect.getargspec. We have to try and parse
    # the __doc__ attribute of the function.
    docstr = obj.__doc__
    args = ''
    
    if docstr:
        items = docstr.split('\n')
        if items:
            func_descr = items[0]
            s = func_descr.replace(obj.__name__,'')
            idx1 = s.find('(')
            idx2 = s.find(')',idx1)
            if idx1 != -1 and idx2 != -1 and (idx2>idx1+1):
                args = s[idx1+1:idx2]
                #wi('\t-Method Arguments:', args)
                for arg in args:
                    yield arg
                
    if args=='':
        yield None


def parseKeyValueString(x):
    """Parse a string in format: 'key1:value1,key2:value2,keyN:valueN'.
    Values assumed to be numeric.
    
    Parameters
    ----------
    x : string
        Required format: 'key:value,key:value,...' or 'key=value,key=value,...'

    Returns
    -------
    dict : {key:value, ...}
    """
    if x is None or x == 'None':
        return {}
    x = x.replace(' ','')
    l = re.split('[=:,]', x)
    return {k.lower():float(v) for k,v in zip(l[::2],l[1::2])}





#!/usr/bin/env python
# import
from __future__ import print_function
## batteries
import os
import sys
import pytest
## 3rd party
import pandas as pd
## package
from MGSIM import Utils
from MGSIM.Commands import Genome_rename as Genome_rename_CMD

# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
def test_help(script_runner):
    ret = script_runner.run('MGSIM', 'genome_rename', '-h')
    assert ret.success

def test_main(script_runner):
    fasta_files = ['Escherichia_coli_K-12_MG1655.fna',
                   'Clostridium_perfringens_ATCC_13124.fna']
    fasta_files = [os.path.join(data_dir, x) for x in fasta_files]
    prefix = os.path.join(data_dir, 'renamed')
    if not os.path.isdir(prefix):
        os.mkdir(prefix)
    ret = script_runner.run('MGSIM', 'genome_rename',
                            '--debug', '--prefix', prefix,
                            *fasta_files)
    assert ret.success

def test_main_ambig(script_runner):
    fasta_files = ['Escherichia_coli_K-12_MG1655_ambig.fna']
    fasta_files = [os.path.join(data_dir, x) for x in fasta_files]
    prefix = os.path.join(data_dir, 'renamed')
    if not os.path.isdir(prefix):
        os.mkdir(prefix)
    ret = script_runner.run('MGSIM', 'genome_rename',
                            '--debug', '--prefix', prefix,
                            *fasta_files)
    assert not ret.success




#!/usr/bin/env python

"""
genome_rename: formatting genome sequences in a multi-fasta file for MGSIM

Usage:
  genome_rename [options] <genome_fasta>...
  genome_rename -h | --help
  genome_rename --version

Options:
  <genome_fasta>  Fasta file(s) containing genome sequences.
                  ('-' if from STDIN)
  --prefix=<p>    Output file prefix. 
                  It must differ from the input file paths.
                  The path will be created if it doesn't exist.
                  [Default: .]
  -a=<a>          Number of ambiguous nucleotides allowed in a genome. 
                  [Default: 0]
  -n=<n>          Number of cpus. 
                  [Default: 1]
  --debug         Debug mode (no multiprocessing).
  -h --help       Show this screen.
  --version       Show version.

Description:
  Reformating the sequence names so that they work with MGSIM.
  The sequence names are reformated to conform to naming rules
  for the downstream analyses.
 
  Output
  ------
  Edited fasta written to STDOUT.

  WARNINGS!
  ---------
  There are no checks that the file is actually in fasta format.

"""

# import
## batteries
from docopt import docopt
import sys,os
import re
import logging
from functools import partial
import multiprocessing as mp
## 3rd party
from Bio import SeqIO
## logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)


def seq_rename(inFile, ambig_cutoff=0, prefix='.'):
    (inFileDir, inFileName) = os.path.split(inFile)
    inFileDir = os.path.abspath(inFileDir)
    prefix = os.path.abspath(prefix)
    outFile = os.path.join(prefix, inFileName)

    if inFileDir == prefix:
        msg = 'ERROR: input and output directories are the same'
        sys.stderr.write(msg + '\n')
        return None

    # regexes
    re0 = re.compile(r'.+ complete genome. ')
    re1 = re.compile(r'\W')
    re2 = re.compile(r'^_*(.*?)_*$')
    re3 = re.compile(r'_*complete_genome')
    re4 = re.compile(r'(.{78}).+')
    ambig_chars = re.compile(r'[RYSWKMBVDH]')

    # iterating through sequence
    ambig_cnt = 0
    with open(outFile, 'w') as outFH:
        for i,record in enumerate(SeqIO.parse(inFile, 'fasta')):
            name = record.name
            name = re0.sub('', name)
            name = re1.sub('_', name)
            name = re2.sub(r'>\1', name)
            name = re3.sub('', name)
            name = re4.sub(r'\1', name)
            name = name.lstrip('>') + '__seq{}'.format(i)
            record.id = name
            record.description = name
            ambig_cnt += len(ambig_chars.findall(str(record.seq.upper())))
            SeqIO.write(record, outFH, 'fasta')

    if ambig_cnt > int(ambig_cutoff):
        msg = 'No. of ambig chars ({}) is > than ambig-cutoff ({})'
        raise ValueError(msg.format(ambig_cnt, ambig_cutoff))
        
    msg = 'File written: {}\n'
    sys.stderr.write(msg.format(outFile))


def main(args):
    # input
    if args['<genome_fasta>'][0] == '-':
        args['<genome_fasta>'] = [x.rstrip() for x in sys.stdin]

    # output
    args['--prefix'] = os.path.abspath(args['--prefix'])
    if not os.path.isdir(args['--prefix']):
        os.makedirs(args['--prefix'])
        
    # rename
    if args['--debug']:
        for f in args['<genome_fasta>']:
            seq_rename(f, ambig_cutoff=args['-a'], prefix=args['--prefix'])
    else:
        p = mp.Pool(int(args['-n']))
        seq_rename_p = partial(seq_rename, ambig_cutoff=args['-a'], prefix=args['--prefix'])
        p.map(seq_rename_p, args['<genome_fasta>'])
    

def opt_parse(args=None):
    if args is None:        
        args = docopt(__doc__, version='0.1')
    else:
        args = docopt(__doc__, version='0.1', argv=args)
    main(args)
   



#!/usr/bin/env python
from setuptools import setup, find_packages
import os
import glob
import numpy

    
# dependencies
install_reqs = [
    'docopt>=0.6.2',
    'numpy>=1.10',
    'pandas>=0.18',
    'scipy>=0.17',
    'configobj>=5.0.6',
    'biopython>=1.68'
]

## install main application
desc = 'Metagenome simulation of multiple synthetic communities'
setup(
    name = 'MGSIM',
    version = '0.1.18',
    description = desc,
    long_description = desc + '\n See README for more information.',
    author = 'Nick Youngblut',
    author_email = 'nyoungb2@gmail.com',
    entry_points={
        'console_scripts': [
            'MGSIM = MGSIM.__main__:main'
        ]
    },
    install_requires = install_reqs,
    include_dirs = [numpy.get_include()],
    license = "MIT license",
    packages = find_packages(),
    package_dir={'MGSIM':
                 'MGSIM'},
    url = 'https://github.com/nick-youngblut/MGSIM'
)







#!/usr/bin/env python
# import
from __future__ import print_function
## batteries
import os
import sys
import pytest
import subprocess
## 3rd party
import numpy as np
import pandas as pd
## package
from MGSIM import Utils
from MGSIM.Commands import Reads as Reads_CMD

# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')

def validate_fastq(data_dir, paired=True):
    # validating fastq
    output_R = os.path.join(data_dir, 'TEST', '1', 'R1.fq')
    ret = subprocess.run(['fqtools', 'validate', output_R])
    assert ret.returncode == 0
    if paired is True:
        output_R = os.path.join(data_dir, 'TEST', '1', 'R2.fq')
        ret = subprocess.run(['fqtools', 'validate', output_R])
        assert ret.returncode == 0

# tests
def test_help(script_runner):
    ret = script_runner.run('MGSIM', 'reads', '-h')
    assert ret.success

def test_main(script_runner):
    genomeList = os.path.join(data_dir, 'genome_list.txt')
    abund_table = os.path.join(data_dir, 'comm_wAbund.txt')
    temp_dir = os.path.join(data_dir, 'temp_read_files')
    output_prefix = os.path.join(data_dir, 'TEST')
    ret = script_runner.run('MGSIM', 'reads',
                            '--art-paired',
                            '--tmp-dir', temp_dir,
                            '--sr-seq-depth', '1e4',
                            '--rndSeed', '8294',
                            genomeList, abund_table, output_prefix)
    assert ret.success
    validate_fastq(data_dir)

def test_main_multiProc(script_runner):
    genomeList = os.path.join(data_dir, 'genome_list.txt')
    abund_table = os.path.join(data_dir, 'comm_wAbund.txt')
    temp_dir = os.path.join(data_dir, 'temp_read_files')
    output_prefix = os.path.join(data_dir, 'TEST')
    ret = script_runner.run('MGSIM', 'reads',
                            '-n', '2',
                            '--art-paired',
                            '--tmp-dir', temp_dir,
                            '--sr-seq-depth', '1e4',
                            '--rndSeed', '8294',
                            genomeList, abund_table, output_prefix)
    assert ret.success
    validate_fastq(data_dir)

def test_main_unpaired(script_runner):
    genomeList = os.path.join(data_dir, 'genome_list.txt')
    abund_table = os.path.join(data_dir, 'comm_wAbund.txt')
    temp_dir = os.path.join(data_dir, 'temp_read_files')
    output_prefix = os.path.join(data_dir, 'TEST')
    ret = script_runner.run('MGSIM', 'reads',
                            '--art-mflen', '0',
                            '--tmp-dir', temp_dir,
                            '--sr-seq-depth', '1e4',
                            '--rndSeed', '8294',
                            genomeList, abund_table, output_prefix)
    assert ret.success
    validate_fastq(data_dir, False)




"""Class for simulating fragments from genomes"""
from __future__ import print_function
# import
## batteries
import os
import sys
import re
import time
import uuid
import logging
import tempfile
import urllib
from urllib.error import HTTPError
from functools import partial
## 3rd party
from multiprocessing import Pool
import pandas as pd
from Bio import SeqIO
from Bio import Entrez

# logging
logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)


# functions
def main(args):
    """Main entry point
    """
    # loading genome accession file
    acc_tbl = parse_acc_table(args['<accession_table>'])    
    
    # creating put directory (if needed)
    if not os.path.isdir(args['-d']):
        os.makedirs(args['-d'])
    
    # query via efetch & download
    logging.info('Downloading genomes via efetch...')
    nprocs = int(float(args['-n']))
    func = partial(e_query, email=args['-e'], outdir=args['-d'],
                   rename=args['-r'], ambig_cutoff=int(args['-a']),
                   tries=int(args['-t']))
    if args['--debug'] is True or nprocs < 2:
        acc_tbl = list(map(func, acc_tbl))
    else:
        pool = Pool(nprocs)
        acc_tbl = pool.map(func, acc_tbl)

    # writing out new acc_tbl to STDOUT
    print('\t'.join(['Taxon', 'Accession', 'Fasta']))
    for x in acc_tbl:
        # writing line
        print('\t'.join([str(y) for y in x]))

def query_assembly(acc, genome_id, outdir, rename=False, ambig_cutoff=0, tries=10):
    """Querying assembly db
    """
    msg = '  Trying to query the "assembly" db with accession: {}'
    logging.info(msg.format(acc))
    
    # query assembly
    record = Entrez.read(Entrez.esearch(db="assembly", term=acc))
    accs = []
    for ID in record['IdList']:
        rec = Entrez.read(Entrez.esummary(db="assembly", id=ID, report="full"), validate=False)
        accession_id = rec['DocumentSummarySet']['DocumentSummary'][0]['AssemblyAccession']
        for attempt in range(tries):
            try:
                refseq_id = Entrez.read(Entrez.esearch(db="nucleotide", term=accession_id), validate=False)['IdList'][0]
                break
            except HTTPError as e:
                logging.info('HTTPError! Error code: {}. Retrying...'.format(e.code))
                time.sleep((attempt + 1) ** 2 + 1)
        accs.append(refseq_id)
    if len(accs) < 1:
        return None
    if len(accs) > 1:
        msg = '  WARNING: >1 refseq accession for assembly: {}'
        logging.warning(msg.format(acc))

    # querying nucleotide
    acc = accs[0]
    records = None
    for attempt in range(tries):
        try:
            genomeIds = Entrez.read(Entrez.esearch(db="nucleotide",
                                                   term=acc,
                                                   retmode="xml",
                                                   retmax=1), validate=False)['IdList']
            records = Entrez.efetch(db="nucleotide",
                                    id=genomeIds,
                                    rettype="fasta",
                                    retmode="text")
        except IOError:
            msg = 'WARNING: Network error! Trying again for accession: {} (attempt {})'
            logging.warning(msg.format(acc, attempt+1))
            records = None
            time.sleep(10 + attempt * 2)
            continue
        if records is not None:
            records = list(records)
            if len(records) == 0 or records[0] == '' or \
               records[0] == 'Supplied id parameter is empty.':
                records = None
            else:
                break
    if records is None:
        sys.exit('Could not download accession: {}'.format(acc))
        
    # writing out file (renaming sequence headers if needed)
    if rename:
        out_file = os.path.join(outdir, '.' + str(uuid.uuid4()) + '.fna')
    else:
        out_file = os.path.join(outdir, genome_id + '.fna')
    with open(out_file, 'w') as outF:
        outF.write(''.join(records) + '\n')

    # checking genome fasta
    out_file = check_genome(out_file, ambig_cutoff)
    
    return out_file
        
def e_query(x, email, outdir, rename=False, ambig_cutoff=0, tries=10):
    """Efetch query for one genome
    x : [taxon, accession]
    """
    genome_id,acc = x[:2]
    logging.info('Downloading genome: {}'.format(genome_id))

    # output fasta file name
    if rename:
        out_file = os.path.join(outdir, '.' + str(uuid.uuid4()) + '.fna')
    else:
        out_file = os.path.join(outdir, genome_id + '.fna')
    
    # querying 
    Entrez.email = email

    ## query nucleotide db
    logging.info('  Querying "nucleotide" db with accession: {}'.format(acc))
    search = acc + '[Accession]'
    records = None
    for attempt in range(tries):
        try:
            genomeIds = Entrez.read(Entrez.esearch(db="nucleotide",
                                                   term=acc,
                                                   retmode="xml",
                                                   retmax=1))['IdList']
            records = Entrez.efetch(db="nucleotide",
                                    id=genomeIds,
                                    rettype="fasta",
                                    retmode="text")
        except IOError:
            msg = 'WARNING: Network error! Trying again for accession: {} (attempt {})'
            logging.warning(msg.format(acc, attempt+1))
            records = None
            time.sleep(10 + attempt * 2)
            continue
        if records is not None:
            records = list(records)
            if len(records) == 0 or records[0] == '' or \
               records[0] == 'Supplied id parameter is empty.':
                records = None
            else:
                break
    if records is None:
        sys.exit('Could not download accession: {}'.format(acc))
        
    # writing out file (renaming sequence headers if needed)
    with open(out_file, 'w') as outF:
        outF.write(''.join(records) + '\n')

    # checking that genome is formatted correctly formatted (and complete)
    out_file = check_genome(out_file, ambig_cutoff)
    if out_file is None:
        out_file = query_assembly(acc, genome_id, outdir, rename=rename,
                                  ambig_cutoff=ambig_cutoff)

    # renaming sequence headers (if needed)    
    if out_file is not None and rename:
        in_file = out_file
        out_file = os.path.join(outdir, genome_id + '.fna')
        genome_rename(in_file, out_file, genome_id)
        
    # status
    msg = '  File written: {}'
    logging.info(msg.format(out_file))
    return [genome_id, acc, out_file]

def check_genome(genome_fasta, ambig_cutoff=0):
    """Checking that genome fasta is formatted correctly
    """
    ambig_chars = re.compile(r'[RYSWKMBVDH]')
    N_cnt = 0
    ambig_cnt = 0
    len_cnt = 0
    for seq_record in SeqIO.parse(genome_fasta, 'fasta'):
        len_cnt += len(seq_record)
        seq = seq_record.seq.upper()
        N_cnt += seq.count('N')
        ambig_cnt += len(ambig_chars.findall(str(seq)))
            
    if ambig_cnt > 0:
        if ambig_cnt > ambig_cutoff:
            msg = '  WARNING: no. of ambig nucleotides is {}; removing genome: {}'
            logging.warning(msg.format(ambig_cnt, genome_fasta))
            return(None)
        else:
            msg = '  WARNING: no. of ambig nucleotides is {}'
            msg += ', but ambig-cutoff is {}, so keeping genome: {}'
            logging.warning(msg.format(ambig_cnt, ambig_cutoff, genome_fasta))            

    if len_cnt < 1000:
        msg = '  WARNING: Genome length < 1000bp; removing genome: {}'
        logging.warning(msg.format(genome_fasta))
        os.remove(genome_fasta)
        return(None)
    if N_cnt / float(len_cnt) > 0.5:
        msg = '  WARNING: Genome sequences are >50% N\'s; removing genome: {}'
        logging.warning(msg.format(genome_fasta))
        os.remove(genome_fasta)
        return(None)

    return genome_fasta

def genome_rename(in_file, out_file, taxon):
    """Renaming sequences in genome file.
    All will be named based on `taxon`
    """
    logging.info('  Renaming genome seqs to: {}'.format(taxon))

    # number of sequences in the genome fasta file
    seq_cnt = 0
    with open(in_file) as inF:
        for line in inF:
            if line.startswith('>'):
                seq_cnt += 1
                
    # renaming sequence headers
    with open(in_file) as inF, open(out_file, 'w') as outF:
        idx = 1
        for line in inF:
            if line.startswith('>'):
                if seq_cnt < 2:
                    outF.write('>{}\n'.format(taxon))
                else:
                    outF.write('>{}_{}\n'.format(taxon, idx))
                    idx += 1
            else:
                outF.write(line)

    # removing temp file
    os.remove(in_file)

def parse_acc_table(infile):
    """Parsing tab-delim accession table (genome_name<tab>accession)
    """
    # load file
    if infile == '-':
        inF = sys.stdin
    else:
        inF = open(infile)
    df = pd.read_csv(inF, sep='\t')
    ## check headers
    diff = set(['Taxon','Accession']) - set(df.columns.values)
    if len(diff) > 0:
        diff = ','.join(diff)
        raise ValueError('Cannot find table columns:{}'.format(diff))
    # removing non-safe characters from taxon labels
    df['Taxon'] = df['Taxon'].apply(lambda x: re.sub('[^A-Za-z0-9-]+', '_', x))
    df['Taxon'] = df['Taxon'].apply(lambda x: re.sub('_+$', '', x))
    # checking for duplicates
    if any(df.duplicated(subset=['Taxon'])):
        raise ValueError('Duplicate taxon names in the accession table')
    
    # convert df to list of lists
    tbl = []
    for i,x in df.iterrows():
        tbl.append([x['Taxon'], x['Accession']])
    return tbl
        
